"""Build script for the C extension."""

import os
import platform
import subprocess
import sys
from typing import Any, MutableMapping, Text

from setuptools import Extension
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):  # type: ignore
    # pylint: disable=too-few-public-methods
    """Cmake extension."""

    def __init__(self, name: Text, sourcedir: Text = '') -> None:
        """Init the extension."""
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


extensions = [  # pylint: disable=invalid-name
    CMakeExtension('snmp_fetch.capi'),
]


class CMakeBuild(build_ext):  # type: ignore
    """Cmake builder."""

    def run(self) -> None:
        """Run the cmake build."""
        try:
            # out = subprocess.check_output(['cmake', '--version'])
            subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError(
                'CMake must be installed to build the following extensions: ' +
                ', '.join(e.name for e in self.extensions)
            )

        if platform.system() == 'Windows':
            raise RuntimeError('Windows is not a supported platform')
            # cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            # if cmake_version < '3.1.0':
            #     raise RuntimeError('CMake >= 3.1.0 is required on Windows')

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext: Extension) -> None:  # type: ignore
        """Build the extension."""
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
            '-DPYTHON_EXECUTABLE=' + sys.executable
        ]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == 'Windows':
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            # build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


def build(setup_kwargs: MutableMapping[Text, Any]) -> None:
    """Build the cmake extension."""
    setup_kwargs.update({
        'ext_modules': extensions,
        'cmdclass': {
            'build_ext': CMakeBuild
        }
    })



"""Python wrapper to the C API."""

from typing import Any, Optional, Sequence, Text, Tuple

import pandas as pd
from toolz.sandbox.core import unzip

from snmp_fetch.capi import PduType, SnmpConfig, SnmpError, SnmpErrorType
from .distributed import distribute
from .distributed import fetch as dfetch
from .distributed import process_response
from .var_bind import VarBind

__all__ = [
    'PduType', 'SnmpConfig', 'SnmpError', 'SnmpErrorType'
]


def fetch(
        pdu_type: PduType,
        df: Any,
        var_binds: Sequence[VarBind],
        config: Optional[SnmpConfig] = None,
        **kwargs: Text
) -> Tuple[Any, Sequence[SnmpError]]:
    """Fetch SNMP results and map to a DataFrame."""
    params = distribute(
        pdu_type,
        df,
        var_binds,
        config,
        batch_size=None,
        **kwargs
    )

    result_dfs, error_lists = unzip([
        process_response(*process_response_params, dfetch(*fetch_params))
        for fetch_params, process_response_params in params
    ])

    return (
        pd.concat(result_dfs),
        [error for errors in error_lists for error in errors]
    )



"""Distributed friendly implementation."""

from functools import reduce
from typing import Any, Optional, Sequence, Text, Tuple, cast

import numpy as np
import pandas as pd

from . import PduType, SnmpConfig, SnmpError
from .capi import fetch as capi_fetch
from .var_bind import VarBind

RESERVED_COL_NAMES = [
    '#oid_size', '#result_size', '#result_type', '#oid', '#ipadding', '#dpadding', '#timestamp',
    '#timestamp'
]

HOST_T = Tuple[int, Text, Text]  # pylint: disable=invalid-name


def merge(a: Any, b: Any) -> Any:
    """Merge two DataFrame results."""
    df = pd.merge(a, b, how='outer', left_index=True, right_index=True)
    df['#timestamp'] = (
        df[['#timestamp_x', '#timestamp_y']]
        .max(axis=1)
        .astype('datetime64[s]')
    )
    return df.drop(columns=['#timestamp_x', '#timestamp_y'])


def prepare_column(arr: np.ndarray, vb: VarBind) -> Any:
    """Convert structured numpy array into DataFrame."""
    view = arr.view(vb.cstruct())
    df = pd.DataFrame.from_records(
        view.tolist(), columns=view.dtype.names
    )
    # fixup dtypes on empty result set
    if vb.header_cstruct.fields is not None:
        for column, dtype in vb.header_cstruct.fields.items():
            df[column] = df[column].astype(dtype[0])
        df = vb.op(df)
    # if not df.index.empty and not isinstance(df.index, pd.core.indexes.range.RangeIndex):
    if df.index.names is not None and [i for i in df.index.names if i is not None]:
        df = df.reset_index().set_index(['#index', *df.index.names])
    else:
        df = df.set_index('#index')
    df = df.drop(columns={
        '#oid_size', '#result_size', '#result_type', '#oid', '#ipadding', '#dpadding'
    }.intersection(df.columns))
    return df


def fetch(
        pdu_type: PduType,
        hosts: Sequence[HOST_T],
        var_binds: Sequence[VarBind],
        config: SnmpConfig
) -> Tuple[Sequence[np.ndarray], Sequence[SnmpError]]:
    """Wrap the C API versions of fetch."""
    return capi_fetch(
        pdu_type,
        hosts,
        [vb() for vb in var_binds],
        config=config if config is not None else SnmpConfig()
    )


def process_response(
        df: Any, var_binds: Sequence[VarBind], idx: Sequence[Text],
        response: Tuple[Sequence[np.ndarray], Sequence[SnmpError]]
) -> Tuple[Any, Sequence[SnmpError]]:
    """Process SNMP response into a results DataFrame and errors list."""
    results, errors = response
    results_df = reduce(
        merge, [prepare_column(r, vb) for r, vb in zip(results, var_binds)]
    )
    results_df['#timestamp'] = results_df['#timestamp'].dt.tz_localize('UTC')
    results_df = (
        results_df.reset_index()
        .set_index('#index')
        .merge(
            df.set_index('#index'), how='outer', left_index=True, right_index=True
        )
        .reset_index(drop=True)
        .set_index(idx)
    )
    return results_df, errors


def distribute(
        pdu_type: PduType,
        df: Any,
        var_binds: Sequence[VarBind],
        config: Optional[SnmpConfig] = None,
        batch_size: Optional[int] = None,
        **kwargs: Text
) -> Sequence[Tuple[
    Tuple[PduType, Sequence[HOST_T], Sequence[VarBind], SnmpConfig],
    Tuple[Any, Sequence[VarBind], Sequence[Text]]
]]:
    """Fetch SNMP results and map to a DataFrame."""
    err_col_names = set([*df.index.names, *df.columns]).intersection(RESERVED_COL_NAMES)
    if err_col_names:
        raise ValueError(
            f'DataFrame contains the following reserved column names: {err_col_names}'
        )

    if config is None:
        config = SnmpConfig()

    host_column = kwargs.pop('host', 'host')
    community_column = kwargs.pop('snmp_community', 'snmp_community')

    idx = df.index.names
    df = df.reset_index()
    df.index = df.index.set_names(['#index'])
    df = df.reset_index()

    def _prepare(df: Any) -> Tuple[
            Tuple[PduType, Sequence[HOST_T], Sequence[VarBind], SnmpConfig],
            Tuple[Any, Sequence[VarBind], Sequence[Text]]
    ]:
        # mypy does not detect that closure of config is no longer optional
        return ((  # type: ignore
            pdu_type,
            [
                (i, str(h), c) for i, h, c
                in df[['#index', host_column, community_column]].values
            ],
            var_binds,
            config
        ), (
            df,
            var_binds,
            cast(Sequence[Text], idx)
        ))

    if batch_size:
        batched = [df[i:i+batch_size] for i in range(0, df.shape[0], batch_size)]
        return [_prepare(batch) for batch in batched]
    return [_prepare(df)]



"""Utility functions."""

from typing import Any, Callable, Text, TypeVar

FuncType = Callable[..., Any]
F = TypeVar('F', bound=FuncType)


def monkeypatch(cls: type, method: Text) -> Callable[[F], F]:
    """Monkey patch and store base method on the function object."""
    def patch(f: F) -> F:
        if not getattr(cls, method) == f:
            setattr(f, method, getattr(cls, method))
            setattr(cls, method, f)
        return f
    return patch



"""Composable DataFrame functions."""

from typing import Any, Callable, List, Optional, Text, Union, cast

import numpy as np
import pandas as pd

from .types.inet_address import IPV4_PREFIX_LOOKUP_TABLE, IPV6_PREFIX_LOOKUP_TABLE, IpNetwork, ip
from .utils import bytes_to_int


def set_index(columns: Union[Text, List[Text]]) -> Callable[[Any], Any]:
    """Return a composable function to set an index on a DataFrame."""
    def _set_index(df: Any) -> Any:
        for column in columns if isinstance(columns, List) else [columns]:
            if column not in df.columns:
                df[column] = None
        return df.set_index(columns)
    return _set_index


def astype(column: Text, dtype: Any) -> Callable[[Any], Any]:
    """Return a composable function to set a dtype of a DataFrame column."""
    def _astype(df: Any) -> Any:
        df[column] = df[column].astype(dtype)
        return df
    return _astype


def decode(column: Text) -> Callable[[Any], Any]:
    """Return a composable function to decode a DataFrame text column."""
    def _decode(df: Any) -> Any:
        df[column] = df[column].str.decode('utf-8', errors='ignore')
        return df
    return _decode


def to_timedelta(
        column: Text, denominator: int = 1, unit: Text = 'seconds'
) -> Callable[[Any], Any]:
    """Return a composable function to convert to a time delta."""
    def _to_timedelta(df: Any) -> Any:
        df[column] = pd.to_timedelta(df[column] // denominator, unit=unit)
        return df
    return _to_timedelta


def drop(**kwargs: Any) -> Any:
    """Return a composable function to drop data from a DataFrame."""
    def _drop(df: Any) -> Any:
        return df.drop(**kwargs)
    return _drop


def to_oid_string(column: Text, size_col: Optional[Text] = '#result_size') -> Callable[[Any], Any]:
    """Return a composable function to convert an uint64 array to an OID string."""
    def _to_oid_string(df: Any) -> Any:
        def arr(x: np.ndarray) -> np.ndarray:
            if size_col is not None:
                return cast(np.ndarray, x[column][:x[size_col] >> 3])
            return cast(np.ndarray, x[column])
        if not df.empty:
            df[column] = (
                df.apply(
                    lambda x: '.' + '.'.join(map(str, arr(x))),
                    axis=1
                )
            )
        return df
    return _to_oid_string


def to_ipv4_address(column: Text) -> Callable[[Any], Any]:
    """Return a composable function to convert a numpy array to an IPv4Address."""
    def _to_ipv4_address(df: Any) -> Any:
        df[column] = df[column].apply(lambda x: ip.IPv4Address(bytes_to_int(x)))
        return df
    return _to_ipv4_address


def to_ipv6_address(column: Text) -> Callable[[Any], Any]:
    """Return a composable function to convert a numpy array to an IPv6Address."""
    def _to_ipv6_address(df: Any) -> Any:
        df[column] = df[column].apply(lambda x: ip.IPv6Address(bytes_to_int(x)))
        return df
    return _to_ipv6_address


def to_cidr_address(
        column: Text, ip_address: Text, mask_or_prefix: Text, **kwargs: Any
) -> Callable[[Any], Any]:
    """Return a composable function to convert an IP address and mask/prefix to a network."""
    def _ip_network(row: Any) -> IpNetwork:
        if not isinstance(row[ip_address], (ip.IPv4Address, ip.IPv6Address)):
            addr = ip.ip_address(row[ip_address])
        else:
            addr = row[ip_address]
        if isinstance(row[mask_or_prefix], int):
            prefix = row[mask_or_prefix]
        elif isinstance(addr, ip.IPv4Address) and isinstance(row[mask_or_prefix], ip.IPv4Address):
            prefix = IPV4_PREFIX_LOOKUP_TABLE[int(row[mask_or_prefix])]
        elif isinstance(addr, ip.IPv6Address) and isinstance(row[mask_or_prefix], ip.IPv6Address):
            prefix = IPV6_PREFIX_LOOKUP_TABLE[int(row[mask_or_prefix])]
        else:
            raise ValueError(
                f'Unable to infer CIDR address from {(row[[ip_address, mask_or_prefix]])}\n'
            )
        return cast(IpNetwork, ip.ip_network((addr, prefix), **kwargs))

    def _to_cidr_address(df: Any) -> Any:
        if df.empty:
            df[column] = None
        else:
            df[column] = df[[ip_address, mask_or_prefix]].apply(_ip_network, axis=1)
        return df
    return _to_cidr_address



"""Helper functions for dealing with variable length data in a numpy array."""

from typing import Any, Callable, List, Optional, Text, Tuple, Type, Union

import numpy as np
from mypy_extensions import Arg, KwArg

from .types.inet_address import ip
from .utils import bytes_to_int

EXTRACT_T = Callable[  # pylint: disable=invalid-name
    [Arg(Any, 'df'), Arg(Text, 'source'), KwArg(Any)], Tuple[Any, np.ndarray]
]
EXTRACT2_T = Callable[  # pylint: disable=invalid-name
    [Arg(Any, 'df'), Arg(Text, 'source'), KwArg(Any)], Tuple[Any, ...]
]
COLUMNS_T = Union[Text, List[Text]]  # pylint: disable=invalid-name
DTYPE_T = Union[Type[object], np.dtype]  # pylint: disable=invalid-name
DTYPES_T = Union[DTYPE_T, List[DTYPE_T]]  # pylint: disable=invalid-name
COMPOSABLE_T = Callable[[Text, Any], Tuple[Text, Any]]  # pylint: disable=invalid-name


def bitstream(source: Text, handler: COMPOSABLE_T) -> Callable[[Any], Any]:
    """Extract values from a structured numpy array with variable length data."""
    def _bitstream(df: Any) -> Any:
        _, df = handler(source, df)
        return df
    return _bitstream


def extract(
        f: EXTRACT_T,
        destination: COLUMNS_T,
        dtypes: Optional[DTYPES_T] = None,
        **kwargs: Any
) -> COMPOSABLE_T:
    """Extract one value from a structured numpy array with variable length data."""
    # flatten return type while enforcing strict input function that returns tail return type of an
    # numpy array
    def expand(f: EXTRACT2_T) -> EXTRACT2_T:
        def _expand(df: Any, source: Text, **kwargs: Any) -> Tuple[Any, ...]:
            results, remaining = f(df, source, **kwargs)
            if isinstance(results, tuple):
                return (*results, remaining)
            return results, remaining
        return _expand

    def _extract(source: Text, df: Any) -> Tuple[Text, Any]:
        if df.empty:
            if isinstance(destination, List) and isinstance(dtypes, List):
                for column, dtype in zip(destination, dtypes):
                    df[column] = None
                    df[column] = df[column].astype(dtype)
            else:
                df[destination] = None
                df[destination] = df[destination].astype(dtypes)
        else:
            columns = [*(destination if isinstance(destination, List) else [destination]), source]
            df[columns] = df.apply(expand(f), axis=1, result_type='expand', source=source, **kwargs)
            if dtypes is not None:
                if isinstance(destination, List) and isinstance(dtypes, List):
                    for column, dtype in zip(destination, dtypes):
                        df[column] = df[column].astype(dtype)
                else:
                    df[destination] = df[destination].astype(dtypes)
        return source, df
    return _extract


def integer(df: Any, source: Text, **kwargs: Any) -> Tuple[int, np.ndarray]:
    # pylint: disable=unused-argument
    """Extract first element of a numpy array as an integer."""
    return df[source][0], df[source][1:]


def object_identifier(df: Any, source: Text, **kwargs: Any) -> Tuple[np.ndarray, np.ndarray]:
    # pylint: disable=unused-argument
    """Extract an object identifier array as a numpy array."""
    return df[source][1:int(df[source][0])+1], df[source][int(df[source][0])+1:]


def inet_address(df: Any, source: Text, **kwargs: Any) -> Tuple[Any, np.ndarray]:
    """Extract an inet address with leading address size and possible zone."""
    default_zone = kwargs.pop('default_zone', None)
    if df[source][0] == 4:
        return (
            (ip.IPv4Address(bytes_to_int(df[source][1:5])), default_zone),
            df[source][5:]
        )
    if df[source][0] == 16:
        return (
            (ip.IPv6Address(bytes_to_int(df[source][1:17])), default_zone),
            df[source][17:]
        )
    if df[source][0] == 8:
        return (
            (ip.IPv4Address(bytes_to_int(df[source][1:9])), bytes_to_int(df[source][9:13])),
            df[source][13:]
        )
    if df[source][0] == 20:
        return (
            (ip.IPv6Address(bytes_to_int(df[source][1:17])), bytes_to_int(df[source][17:21])),
            df[source][21:]
        )
    raise TypeError('Datatype not understood')



"""Enhanced IP address types.

Monkey patches the ipaddress module types to allow for IPv4 and IPv6 comparisons.
"""

import ipaddress as ip
from typing import TypeVar, Union

from snmp_fetch.utils import monkeypatch

T = TypeVar('T')


@monkeypatch(ip.IPv4Address, '__lt__')
def ipv4_address__lt__(self: T, other: T) -> bool:
    # pylint: disable=no-member
    """Compare less than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return True
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__le__')
def ipv4_address__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return True
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__gt__')
def ipv4_address__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return False
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Address, '__ge__')
def ipv4_address__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return ipv4_address__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Address):
        return False
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__lt__')
def ipv6_address__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return False
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__le__')
def ipv6_address__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return False
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__le__.__le__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__gt__')
def ipv6_address__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return True
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv6Address, '__ge__')
def ipv6_address__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP address object types."""
    if isinstance(other, ip.IPv4Address):
        return True
    if isinstance(other, ip.IPv6Address):
        return ipv6_address__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Address or IPv6Address object')


@monkeypatch(ip.IPv4Network, '__lt__')
def ipv4_network__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return True
    raise TypeError(f'{other} is not an IPv4Network or IPv6Network object')


@monkeypatch(ip.IPv4Network, '__le__')
def ipv4_network__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return True
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Network, '__gt__')
def ipv4_network__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return False
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Network, '__ge__')
def ipv4_network__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return ipv4_network__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Network):
        return False
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__lt__')
def ip6_network__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return False
    if isinstance(other, ip.IPv6Network):
        return ip6_network__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__le__')
def ipv6_network__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return False
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__le__.__le__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__gt__')
def ipv6_network__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return True
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv6Network, '__ge__')
def ipv6_network__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP network object types."""
    if isinstance(other, ip.IPv4Network):
        return True
    if isinstance(other, ip.IPv6Network):
        return ipv6_network__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(f'{other} is not an IPv4Network or IPv6network object')


@monkeypatch(ip.IPv4Interface, '__lt__')
def ipv4_interface__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__lt__.__lt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return True
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__le__')
def ipv4_interface__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__le__.__le__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return True
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__gt__')
def ipv4_interface__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__gt__.__gt__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return False
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv4Interface, '__ge__')
def ipv4_interface__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return ipv4_interface__ge__.__ge__(self, other)  # type: ignore
    if isinstance(other, ip.IPv6Interface):
        return False
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__lt__')
def ipv6_interface__lt__(self: T, other: T) -> bool:
    """Compare less than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return False
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__lt__.__lt__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__le__')
def ipv6_interface__le__(self: T, other: T) -> bool:
    """Compare less than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return False
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__le__.__le__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__gt__')
def ipv6_interface__gt__(self: T, other: T) -> bool:
    """Compare greater than on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return True
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__gt__.__gt__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


@monkeypatch(ip.IPv6Interface, '__ge__')
def ipv6_interface__ge__(self: T, other: T) -> bool:
    """Compare greater than or equal to on mixed IP interface object types."""
    if isinstance(other, ip.IPv4Interface):
        return True
    if isinstance(other, ip.IPv6Interface):
        return ipv6_interface__ge__.__ge__(self, other)  # type: ignore
    raise TypeError(
        f'{other} is not an IPv4Interface or IPv6interface object'
    )


IpAddress = Union[ip.IPv4Address, ip.IPv6Address]
IpNetwork = Union[ip.IPv4Network, ip.IPv6Network]
IpInterface = Union[ip.IPv4Interface, ip.IPv6Interface]

IPV4_PREFIX_LOOKUP_TABLE = {
    (0xFFFFFFFF << i) & 0xFFFFFFFF: 32 - i for i in range(0, 33)
}

IPV6_PREFIX_LOOKUP_TABLE = {
    (
        (0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF << i) &
        0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF
    ): 128 - i
    for i in range(0, 129)
}


__all__ = [
    'ip'
]



"""Variable bindings."""

import re
from functools import reduce
from operator import add
from typing import Any, Callable, Mapping, Optional, Sequence, Text, Tuple, Union, cast, overload

import attr
import numpy as np
from toolz.functoolz import compose, identity

from .fp import curry2
from .fp.either import Either, Left, Right
from .fp.maybe import Just, Maybe, Nothing

DTYPE_FIELDS_T = (  # pylint: disable=invalid-name
    Mapping[str, Union[Tuple[np.dtype, int], Tuple[np.dtype, int, Any]]]
)


@overload
def convert_oid(oid: Text) -> Sequence[int]:
    # pylint: disable=unused-argument
    # pragma: no cover
    """Convert a text oid to a sequence of integers."""
    ...  # pragma: no cover


@overload
def convert_oid(oid: Sequence[int]) -> Text:
    # pylint: disable=function-redefined, unused-argument
    # pragma: no cover
    """Convert a sequence of integers to a text oid."""
    ...  # pragma: no cover


def convert_oid(
        oid: Union[Text, Sequence[int]]
) -> Union[Sequence[int], Text]:
    # pylint: disable=function-redefined
    """Convert an oid between text and sequence of integers."""
    if isinstance(oid, Text):
        if re.match(r'^\.?\d+(\.\d+)*$', oid):
            if oid.startswith('.'):
                oid = oid[1:]
            return [int(x) for x in (oid).split('.')]
        raise ValueError(f'{oid} is not a valid oid')
    return '.'+'.'.join(map(str, oid))


@overload
def validate_oid(oid: Text) -> Text:
    # pylint: disable=unused-argument
    # pragma: no cover
    """Validate a text oid."""
    ...  # pragma: no cover


@overload
def validate_oid(oid: Sequence[int]) -> Sequence[int]:
    # pylint: disable=function-redefined, unused-argument
    """Validate a sequence of integers oid."""
    ...  # pragma: no cover


def validate_oid(
        oid: Union[Text, Sequence[int]]
) -> Union[Text, Sequence[int]]:
    """Validate an oid."""
    # pylint: disable=function-redefined
    return convert_oid(convert_oid(oid))


def get_fields(d: np.dtype) -> Either[Exception, DTYPE_FIELDS_T]:
    """Get structured dtype fields safely."""
    if d.fields is None:
        return Left(ValueError(f'structured dtype required: {d}'))
    return Right(d.fields)


def dtype_concat(ds: Sequence[np.dtype]) -> np.dtype:
    """Concat structured datatypes."""
    def _concat(
            acc: Tuple[Mapping[Any, Any], int], a: np.dtype
    ) -> Tuple[DTYPE_FIELDS_T, int]:
        acc_fields, acc_itemsize = acc
        fields = get_fields(a).throw()
        intersection = set(acc_fields).intersection(set(fields))
        if intersection != set():
            raise ValueError(f'dtypes have overlapping fields: {intersection}')
        return (
            {
                **acc_fields,
                **{k: (d[0], d[1] + acc_itemsize) for k, d in fields.items()}
            },
            acc_itemsize + a.itemsize
        )
    # dtype.fields() doesn't match dtype constructor despite being compatible
    return np.dtype(reduce(_concat, ds, (cast(DTYPE_FIELDS_T, {}), 0))[0])  # type: ignore


def dtype_array(d: np.dtype, size: int) -> Maybe[np.dtype]:
    """Return an (n,) datatype if possible.

    Returns a bare datatype for a scalar request or Nothing with a negative size.
    """
    if size <= 0:
        return Nothing()
    if size == 1:
        return Just(d)
    return Just(np.dtype((d, size)))


def maybe_oid(value: Optional[Text]) -> Maybe[Text]:
    """Convert and validate Optional[Text] representation of OID to Maybe[Text]."""
    return Maybe.from_optional(value).fmap(validate_oid)


def maybe_dtype(value: Optional[np.dtype]) -> Maybe[np.dtype]:
    """Convert and validate Optional[dtype] to Maybe[dtype]."""
    if value is not None:
        if value.fields is None:
            raise ValueError(f'structured dtype required: {value}')
        if value.itemsize % 8 != 0:
            raise ValueError(f'dtype required to be 8 bytes aligned: {value}')
    return Maybe.from_optional(value)


@attr.s(frozen=True, slots=True)
class VarBind:
    """Variable binding."""

    oid: Maybe[Text] = attr.ib(
        default=None, converter=maybe_oid
    )
    index: Maybe[np.dtype] = attr.ib(
        default=None, converter=maybe_dtype
    )
    data: Maybe[np.dtype] = attr.ib(
        default=None, converter=maybe_dtype
    )
    op: Callable[[Any], Any] = attr.ib(
        default=identity
    )

    def __lshift__(self, other: 'VarBind') -> 'VarBind':
        # pylint: disable=no-member
        """Combine variable bindings."""
        return VarBind(
            oid=(
                self.oid
                .combine(curry2(add), other.oid)
                .to_optional()
            ),
            index=(
                self.index
                .combine(
                    lambda x: lambda y: dtype_concat([x, y]),
                    other.index
                )
                .to_optional()
            ),
            data=(
                self.data
                .combine(
                    lambda x: lambda y: dtype_concat([x, y]),
                    other.data
                )
                .to_optional()
            ),
            op=compose(other.op, self.op)
        )

    header_cstruct = np.dtype([
        ('#index', np.uint64),
        ('#oid_size', np.uint64),
        ('#result_size', np.uint64),
        ('#result_type', np.uint64),
        ('#timestamp', 'datetime64[s]')
    ])

    @property
    def oid_cstruct(self) -> np.dtype:
        # pylint: disable=no-member
        """Get the oid cstruct."""
        return (
            self.oid
            .bind(lambda x: (
                dtype_array(np.dtype(np.uint64), len(convert_oid(x)))
                .fmap(lambda arr: np.dtype([('#oid', arr)]))
            ))
            .fail(ValueError('oid has no dtype'))
        )

    def cstruct(self) -> np.dtype:
        # pylint: disable=no-member
        """Create the var_bind cstruct."""
        return dtype_concat([
            self.header_cstruct,
            self.oid_cstruct,
            *Maybe.cat([
                self.index,
                self.data
            ])
        ])

    def __call__(
            self, param: Optional[Text] = None
    ) -> Tuple[Sequence[int], Tuple[int, int]]:
        """Return a null variable binding cstruct with optional parameter."""
        return convert_oid(
            self.oid
            .combine(curry2(add), maybe_oid(param))
            .fail(AttributeError('var_bind has no oid'))
        ), (
            self.oid_cstruct.itemsize + (
                self.index
                .fmap(lambda x: x.itemsize)
                .from_maybe(0)
            ), (
                self.data
                .fmap(lambda x: x.itemsize)
                .from_maybe(0)
            )
        )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['snmp_fetch', 'snmp_fetch.df', 'snmp_fetch.df.types', 'snmp_fetch.fp']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.1,<20.0',
 'numpy>=1.16,<2.0',
 'pandas>=0.25,<0.26',
 'toolz>=0.10.0,<0.11.0']

extras_require = \
{'notebooks': ['jupyterlab>=1.1,<2.0', 'distributed>=2.6,<3.0']}

setup_kwargs = {
    'name': 'snmp-fetch',
    'version': '0.1.7.1',
    'description': 'An opinionated python SNMPv2 library built for rapid database ingestion.',
    'long_description': 'snmp-fetch\n==========\n\n|Version badge| |Python version badge| |PyPI format badge| |Build badge| |Coverage badge|\n\n.. |Version badge| image:: https://img.shields.io/pypi/v/snmp-fetch\n   :target: https://pypi.org/project/snmp-fetch/\n\n.. |Python version badge| image:: https://img.shields.io/pypi/pyversions/snmp-fetch\n   :alt: PyPI - Python Version\n   :target: https://pypi.org/project/snmp-fetch/\n  \n.. |PyPI format badge| image:: https://img.shields.io/pypi/format/snmp-fetch\n   :alt: PyPI - Format\n   :target: https://pypi.org/project/snmp-fetch/\n\n.. |Build badge| image:: https://travis-ci.org/higherorderfunctor/snmp-fetch.svg?branch=master\n   :target: https://travis-ci.org/higherorderfunctor/snmp-fetch\n\n.. |Coverage badge| image:: https://coveralls.io/repos/github/higherorderfunctor/snmp-fetch/badge.svg\n   :target: https://coveralls.io/github/higherorderfunctor/snmp-fetch\n\nAn opinionated python3.7 SNMPv2 package designed for rapid database ingestion.  This package is a source distribution that includes a C module wrapping net-snmp.  No MIB processing is done as part of this package.  The C module copies raw results from net-snmp into numpy arrays for fast post-processing with either numpy or pandas.  Other libraries that wrap net-snmp will typically return control to python between every PDU request-response.  Snmp-fetch is designed to be thread-safe and efficient by walking multiple targets within the C module with the GIL released.  Helper modules are provided to aid in the post-processing with MIB-like definitions for converting the raw data into usable DataFrames.\n\nPrerequisites\n"""""""""""""\n\nSnmp-fetch requires python 3.7, a c++17 compiler (currently only supports gcc-8), and cmake 3.12.4+.  No other user installed dependencies should be required for building this package.\n\n.. ATTENTION::\n\n   Installation can take awhile as the install script will download boost and download and build a light-weight version of net-snmp 5.8 within the package.\n\n   The boost download can take awhile as it clones each submodule as oppose to downloading the compressed distribution.  There is an issue with downloading the compressed distribution via cmake with 302 redirects to a file failing in cURL.\n\n   The cmake script will attempt to detect the number of cores on the host machine to speedup download and build times.  Expect installation times to range from 5 minutes (4 cores with hyperthreading) to 30+ minutes (1 core).\n\nInstallation\n""""""""""""\n\n.. code:: console\n\n   # poetry\n   poetry add snmp-fetch --no-dev\n   # pip\n   pip install snmp-fetch\n\nExamples\n""""""""\n\nThe examples use jupyter and the dependencies can be installed using the following:\n\n.. code:: console\n\n   git clone --recurse-submodules -j8 https://github.com/higherorderfunctor/snmp-fetch.git\n   cd snmp_fetch\n   virtualenv -p python3.7 ENV\n   source ENV/bin/activate\n   poetry install -E notebooks\n   jupyter lab\n\nDevelopment\n"""""""""""\n\n`Poetry <https://poetry.eustace.io/>`_ is required for the development of snmp-fetch.\n\n.. code:: console\n\n   # clone the respository\n   git clone --recurse-submodules -j8 https://github.com/higherorderfunctor/snmp-fetch.git\n   cd snmp-fetch\n\n   # if working off an existing clone, update the current branch\n   git pull  # pull the latest code\n   git submodule update --init --recursive  # pull the latest submodule version\n\n   # setup the virtual environment - mypy uses symbolic links in the \'stubs\' directory to\n   # expose packages that play nicely with the static type checker\n   virtualenv -p python3.7 ENV\n   source ENV/bin/activate\n   poetry install\n\n.. code:: console\n\n   # C++ headers are in the following folders for linters\n   export CPLUS_INCLUDE_PATH="build/temp.linux-x86_64-3.7/include:lib/pybind11/include:lib/Catch2/single_include/catch2"\n\n   # python linting\n   poetry run isort -rc --atomic .\n   poetry run pylint snmp_fetch tests\n   poetry run flake8 snmp_fetch tests\n   poetry run mypy -p snmp_fetch -p tests\n   poetry run bandit -r snmp_fetch\n\n   # C++ linting\n   # TODO\n\n   # python testing\n   poetry run pytest -v --hypothesis-show-statistics tests\n   # fail fast testing\n   poetry run pytest -x --ff tests\n\n   # C++ testing\n   pushd build/temp.linux-x86_64-3.7/\n   cmake -DBUILD_TESTING=ON ../.. && make test_capi test\n   popd\n\n\nKnown Limitations\n"""""""""""""""""\n- Changes between v0.1.x versions may introduce breaking changes.\n\n- The library only supports SNMPv2 at this time.\n\n- `BULKGET_REQUEST` and `NEXT_REQUEST` will always perform a walk.\n\n- Walks will always end if the root of the OID runs past the requested OID.\n\n- Duplicate objects on the same host/request will be silently discarded.\n\n  - This includes the initial request; walks must be performed on an OID prior to the first desired.\n\n- NO_SUCH_INSTANCE, NO_SUCH_OBJECT, and END_OF_MIB_VIEW response variable bindings are exposed as errors for handling by the client.\n',
    'author': 'Christopher Aubut',
    'author_email': 'christopher@aubut.me',
    'url': 'https://github.com/higherorderfunctor/snmp-fetch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



"""Auxiliary helper functions for DataFrame processing."""

import numpy as np


def bytes_to_int(x: np.ndarray) -> int:
    """Convert bytes to a single integer."""
    return int.from_bytes(x.astype(np.uint8).tobytes(), byteorder='big')
